<?php

require_once(dirname(__FILE__) . '/rfc2047.php');

class rfc2822Address
{
	var $m_ShouldBeEncoded	= false;	// Whether or not this address' display-name needs to be encoded per RFC 2047
	var $m_displayName		= "";		// display-name
	var $m_addrSpec			= "";		// addr-spec

	public function setShouldBeEncoded($shouldEncode = false)
	{
		$this->m_ShouldBeEncoded = $shouldEncode;
	}

	public function getShouldBeEncoded()
	{
		return $this->m_ShouldBeEncoded;
	}

	public function setDisplayName($displayName)
	{
		$this->m_displayName = $displayName;
	}

	public function getDisplayName()
	{
		return $this->m_displayName;
	}

	public function setAddrSpec($addrSpec)
	{
		$this->m_addrSpec = $addrSpec;
	}

	public function getAddrSpec()
	{
		return $this->m_addrSpec;
	}
}

class rfc2822AddressList
{
	var $m_RawContents			= "";		// The string representation of the address-list
	var $m_LineLength			= 0;		// The length of the current (appendable) line
	var $m_LineHasEncodedWord	= false;	// Whether or not the current line has an encoded-word on it.
	var $m_Charset				= "";		// The character set of any encoded data
	var $m_VirginLine			= true;		// Whether or not any data has been placed on the current line.

	public function __construct($charset)
	{
		$this->m_Charset = $charset;
	}

	// Called to append an item to the address list
	// Input is an rfc2822Address object
	public function appendItem($item)
	{
		if (!$item->getShouldBeEncoded())
		{
			// No encoding necessary.
			$lineLimit = ($this->m_LineHasEncodedWord ? rfc2047::HEADER_LINE_SHORT_LIMIT : rfc2047::HEADER_LINE_LONG_LIMIT);
			$lineLeft  = ($lineLimit - $this->m_LineLength);
			$toAppend  = "";

			if ("" != $item->getDisplayName())
			{
				$toAppend = '"' . $item->getDisplayName() . '" <' . $item->getAddrSpec() . '>';			
			}
			else
			{
				$toAppend = $item->getAddrSpec();
			}
			
			$toAppendLen = strlen($toAppend);
			
			if (($toAppendLen + 3) <= $lineLeft)
			{
				$this->myStartNewItem(false);
				$this->myAppendRaw($toAppend);
			}
			else
			{
				$this->myStartNewItem(true);
				$this->myAppendRaw($toAppend);
			}
		}
		else
		{
			// Encoding is required.
			$totalBytes   = strlen($item->getDisplayName());
			$doneBytes    = 0;
			$addrSpec	  = '<' . $item->getAddrSpec() . '>';
			$isFirstChunk = true;
			
			do
			{
				$lineLeft     = (rfc2047::HEADER_LINE_SHORT_LIMIT - $this->m_LineLength);
				$bytesFitHere = $this->myBytesFitOnThisLine($lineLeft, $addrSpec);

				if ($bytesFitHere < ($totalBytes - $doneBytes))
				{
					if (!$bytesFitHere)
					{
						// No more can fit on this line.
						if ($isFirstChunk)
						{
							$this->myStartNewItem(true);
						}
						else
						{
							$this->myFold();
						}
					}
					else
					{
						// Some can fit here.
						$doneBytes += $this->myProcessEncodedChunk($item, $doneBytes, $bytesFitHere, $isFirstChunk);
					}
				}
				else
				{
					// The remainder will fit on this line.
					$doneBytes += $this->myProcessEncodedChunk($item, $doneBytes, $bytesFitHere, $isFirstChunk);
					$this->myAppendRaw(' ' . $addrSpec, true);
				}
				
			} while ($doneBytes < $totalBytes);
		}
	}

	private function myProcessEncodedChunk($item, $doneBytes, $bytesFitHere, &$first)
	{
		$thisChunk = mb_strcut($item->getDisplayName(), $doneBytes, $bytesFitHere, $this->m_Charset);
		
		if ($first)
		{
			$this->myStartNewItem(false);
			$first = false;
		}
		
		$this->myAppendRaw(rfc2047::getEncodedForm($thisChunk, $this->m_Charset), true);
		return strlen($thisChunk);
	}
	
	private function myBytesFitOnThisLine($lineLeft, $addrSpec)
	{
		$realLineLeft = ($lineLeft - (strlen($addrSpec) + ($this->m_VirginLine ? 2 : 4)));
		return rfc2047::getMaximumEncodeBytes($this->m_Charset, $realLineLeft);
	}
	
	private function myStartNewItem($newLine)
	{
		if ($newLine)
		{
			$this->m_RawContents .= ',';
			$this->myFold();
		}
		else
		{
			// Only append if the line is not virgin
			if (!$this->m_VirginLine)
			{
				$this->myAppendRaw(', ');
			}
		}
	}
	
	private function myFold()
	{
		$this->m_RawContents       .= rfc2047::HEADER_LINEFEED;
		$this->m_LineLength         = 1;
		$this->m_VirginLine         = true;
		$this->m_LineHasEncodedWord = false;
	}
	
	private function myAppendRaw($str, $encoded = false)
	{
		$this->m_RawContents       .= $str;
		$this->m_LineLength        += strlen($str);
		$this->m_VirginLine         = false;
		
		if ($this->m_LineHasEncodedWord && !$encoded)
			return;	// Don't un-set the encoded flag if it's already set.

		$this->m_LineHasEncodedWord = $encoded;
	}

	public function setCharset($charset)
	{
		$this->m_Charset = $charset;
	}

	public function getRawContents()
	{
		return $this->m_RawContents;
	}
}

function rfc2822_encodeAddressList($input, $charset = 'ISO-8859-1')
{
	$retVal = new rfc2822AddressList($charset);
	
	// Notes:
	//
	// - Input is a comma-delimited list of RFC 2822-compliant address fields, such as:
	//	Name <user@domain.tld>
	//	user@domain.tld
	//
	// - If any name is encoded, the line it appears on is limited to 76 characters.
	// - Header "folding" per RFC 2822 must only take place directly after the high-level syntactic break (the comma).
	// - encoded-words must not appear inside quoted-strings, so if the name is encoded, the quotes should be ignored.

	if (!preg_match_all('@"[^"]*"[ ]<[-\\w!#$%&\'*+/=?^_`{|}~.]+\\@[-._\\w]+>|[-!#$%&\'*+/=?^_`{|}~\\w ]+[ ]<[-\\w!#$%&\'*+/=?^_`{|}~.]+\\@[-._\\w]+>|[-\\w!#$%&\'*+/=?^_`{|}~.]+\\@[-._\\w]+@i', $input, $matches))
	{
		// No input?
		return $retVal->getRawContents();
	}

	// Got at least one address to parse and possibly encode. Walk the list of addresses and handle them
	// one at a time.
	foreach ($matches[0] as $addr)
	{
		$addrObj  = new rfc2822Address();
		$addr	    = trim($addr);
		
		// What kind of address format is this one? If it has a name part that includes
		// any non-ASCII characters, encode it.
		if (preg_match('@^["]?(.+?)["]?[\s]+<(.+)>$@', $addr, $matches))
		{
			// Looks like the full form of the address (["]name["] <addr>). The name ($matches[1]) could possibly
			// require encoding.
			if (preg_match('@[\x80-\xFF]@', $matches[1]))
			{
				// The name contains bytes that are not 7-bit ASCII, so we need to encode it
				$addrObj->setShouldBeEncoded(true);
			}

			$addrObj->setDisplayName($matches[1]);
			$addrObj->setAddrSpec($matches[2]);
		}
		else
		{
			// If there's no match, it's probably just the addr-spec, and there's nothing to encode.
			// This form of the address is atomic and can't be reduced.	
			$addrObj->setAddrSpec($addr);
		}

		$retVal->appendItem($addrObj);
	}
	
	return $retVal->getRawContents();
}

function rfc2822_encodeUnstructuredHeader($input, $charset = 'ISO-8859-1')
{
	// If the header doesn't contain any non-7-bit ASCII bytes, it isn't necessary to encode it.
	if (!preg_match('@[\x80-\xFF]@', $input))
	{
		return $input;
	}
	
    $retVal = "";
	
	// Encode as much as possible at a time.
	$totChunkSize = strlen($input);
	$maxChunkSize = rfc2047::getMaximumEncodeBytes($charset);

	// If we can encode all of the input at once, do that (there is a limit on the size of an encoded-word)
	// If we can't, split it up into chunks.  
	if ($totChunkSize <= $maxChunkSize)
	{
		// The whole input will fit in an encoded-word
		$toEncode = array($input);
	}
	else
	{
		// The input needs to be split up into multiple encoded-words.
		$toEncode = rfc2822_mbSafeSplitString($input, $maxChunkSize, $charset);
	}
	
	// Process each chunk that we need to encode as an encoded-word,
	// dropping to a new line when necessary.
	$lineCntr = 0;

	foreach ($toEncode as $curChunk)
	{
        $encodedWordSize = rfc2047::getEncodedWordSize($curChunk, $charset);
        $encodedWord     = rfc2047::getEncodedForm($curChunk, $charset);
        
        if ($encodedWordSize <= (rfc2047::HEADER_LINE_SHORT_LIMIT - $lineCntr))
        {
            // We can fit this on the current line.
            $retVal   .= $encodedWord;
            $lineCntr += $encodedWordSize;
        }
        else
        {
            // Can't fit it on this line, drop.
            $retVal   .= rfc2047::HEADER_LINEFEED;
            $retVal   .= $encodedWord;
            $lineCntr = 1 + $encodedWordSize; // 1 is for the space that begins the new line.
        }
	}

    return $retVal;
}

/**
* Splits a string into an array of n-sized chunks of bytes.
* If splitting the string on a byte boundary that would break encoding such as UTF-8, the chunk is rounded down
* to the next character boundary.
* RML
*/
function rfc2822_mbSafeSplitString($input, $chunksize, $charset)
{
    $inputLen = strlen($input); // Number of bytes in the input string
    $doneLen  = 0;              // How many bytes processed
    $ret      = array();

    do
    {
        $chunk = mb_strcut($input, $doneLen, $chunksize, $charset);
        $bytes = strlen($chunk);

        $ret[] = $chunk;
        $doneLen += $bytes;
        
    } while ($doneLen < $inputLen);

    return $ret;
}
    
?>
